@@ -1,6 +1,6 @@
 /* vi:set sw=4:
 
-    mails.c  (TMail version 0.8.15)
+    mails.c  (TMail version 0.8.16)
 
     Copyright (c) 1998-2000 Minero Aoki <aamine@dp.u-netsurf.ne.jp>
 
@@ -645,13 +645,24 @@ intn(str)
 void
 Init_mails()
 {
-    MailScanner = rb_define_class("MailScanner", rb_cObject);
-    rb_define_singleton_method(MailScanner, "new", mails_s_new, 3) ;
+    VALUE TMail;
+    ID sym_tmail = rb_intern("TMail");
+    ID sym_scerr = rb_intern("ScanError");
+
+    if (rb_const_defined(rb_cObject, sym_tmail))
+        TMail = rb_const_get(rb_cObject, sym_tmail);
+    else
+        TMail = rb_define_module("TMail");
+    MailScanner = rb_define_class_under(TMail, "MailScanner", rb_cObject);
+    rb_define_singleton_method(MailScanner, "new", mails_s_new, 3);
     rb_define_method(MailScanner, "scan", mails_scan, 1);
     rb_define_method(MailScanner, "debug", mails_debug_get, 0);
     rb_define_method(MailScanner, "debug=", mails_debug_set, 1);
 
-    ScanError = rb_eval_string("::ScanError");
+    if (rb_const_defined(rb_cObject, sym_scerr))
+        ScanError = rb_const_get(rb_cObject, sym_scerr);
+    else
+        ScanError = rb_define_class("ScanError", rb_eStandardError);
 
     tok_atom   = intn("ATOM");
     tok_digit  = intn("DIGIT");
