@@ -20,23 +20,30 @@
 #include "re.h"
 
 
+#define STRSCAN_VERSION "0.6.0"
+
 struct strscanner
 {
-    char *pbeg;
-    char *ptr;
-    char *pend;
-    char *pre;
-    unsigned int flags;
+    VALUE str;
+    long idx;
+    long pre_beg;
+    long pre_end;
     struct re_registers regi;
 };
 
-#define MATCH_FLAG (1U << 0)
-#define DUP_FLAG   (1U << 1)
+#define S_PTR(s)  (RSTRING(s->str)->ptr)
+#define S_LEN(s)  (RSTRING(s->str)->len)
+#define S_END(s)  (S_PTR(s) + S_LEN(s))
+#define CURPTR(s) (S_PTR(s) + s->idx)
+#define PREPTR(s) (S_PTR(s) + s->pre_beg)
 
-#define MATCHED_P(p) (p->flags & MATCH_FLAG)
+#define CLEAR_MATCH_STATUS(s) s->pre_beg = s->pre_end = -1
+#define SAVE_MATCH_STATUS(s,len) s->pre_beg = s->idx; \
+                                 s->pre_end = s->idx + len
+#define MATCHED_P(s) (s->pre_beg >= 0)
 
 #define GET_SCANNER(obj,dat) Data_Get_Struct(obj, struct strscanner, dat)
-#define SCAN_FINISHED(dat) ((dat)->ptr >= (dat)->pend)
+#define SCAN_FINISHED(dat) ((dat)->idx >= RSTRING(dat->str)->len)
 
 static VALUE StringScanner;
 static VALUE ScanError;
@@ -45,17 +52,22 @@ static VALUE ScanError;
 /* ------------------------------------------------------------- */
 
 
+static void
+strscan_mark(p)
+    struct strscanner *p;
+{
+    rb_gc_mark(p->str);
+}
+
 static void
 strscan_free(p)
     struct strscanner *p;
 {
-    if (p->flags & DUP_FLAG)
-        free(p->pbeg);
     ruby_re_free_registers(&(p->regi));
+    memset(p, sizeof(struct strscanner), 0);
     free(p);
 }
 
-
 static VALUE
 strscan_s_new(argc, argv, klass)
     int argc;
@@ -69,20 +81,13 @@ strscan_s_new(argc, argv, klass)
     Check_Type(str, T_STRING);
 
     p = ALLOC_N(struct strscanner, 1);
-    p->flags = 0U;
-    if (rb_safe_level() >= 4 || RTEST(dup_p)) {
-        p->flags |= DUP_FLAG;
-        p->ptr = ALLOC_N(char, RSTRING(str)->len) ;
-        memcpy(p->ptr, RSTRING(str)->ptr, sizeof(char) * RSTRING(str)->len);
-    } else {
-        p->ptr = RSTRING(str)->ptr;
-    }
-    p->pbeg = p->ptr;
-    p->pend = p->ptr + RSTRING(str)->len;
-    p->pre = 0;
+    p->str = RTEST(dup_p) ? rb_str_dup(str) : str;
+    rb_obj_freeze(p->str);
+    p->idx = 0;
+    CLEAR_MATCH_STATUS(p);
     MEMZERO(&(p->regi), struct re_registers, 1);
 
-    return Data_Wrap_Struct(klass, 0, strscan_free, p);
+    return Data_Wrap_Struct(klass, strscan_mark, strscan_free, p);
 }
 
 
@@ -93,181 +98,200 @@ strscan_reset(self)
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    p->ptr = p->pbeg;
-    p->pre = 0;
-    p->flags &= ~MATCH_FLAG;
+    p->idx = 0;
+    CLEAR_MATCH_STATUS(p);
     return self;
 }
 
 
 static VALUE
-strscan_clear(self)
+strscan_terminate(self)
     VALUE self;
 {
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    p->ptr = p->pend;
-    p->pre = 0;
-    p->flags &= ~MATCH_FLAG;
+    p->idx = S_LEN(p);
+    CLEAR_MATCH_STATUS(p);
     return self;
 }
 
+static VALUE
+strscan_get_string(self)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    return p->str;
+}
 
 static VALUE
-strscan_do_scan(self, regex, ptrflag, strflag)
+strscan_set_string(self, str)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    Check_Type(str, T_STRING);
+    p->str = rb_str_dup(str);
+    rb_obj_freeze(p->str);
+    p->idx = 0;
+    CLEAR_MATCH_STATUS(p);
+    return str;
+}
+
+static VALUE
+strscan_get_pointer(self)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    return INT2FIX(p->idx);
+}
+
+static VALUE
+strscan_set_pointer(self, v)
+    VALUE self, v;
+{
+    struct strscanner *p;
+    long i;
+
+    GET_SCANNER(self, p);
+    i = NUM2INT(v);
+    if (i < 0) i += S_LEN(p);
+    if (i < 0) rb_raise(rb_eRangeError, "index out of range");
+    if (i > S_LEN(p)) rb_raise(rb_eRangeError, "index out of range");
+    p->idx = i;
+    return INT2FIX(i);
+}
+
+
+
+static VALUE
+strscan_do_scan(self, regex, succptr, getstr, headonly)
     VALUE self, regex;
-    int ptrflag, strflag;
+    int succptr, getstr, headonly;
 {
     struct strscanner *p;
     int ret;
+    char *pbeg;
+    long plen;
 
     Check_Type(regex, T_REGEXP);
     GET_SCANNER(self, p);
 
-    p->flags &= ~MATCH_FLAG;
-    ret = re_match(RREGEXP(regex)->ptr,
-                   p->ptr,
-                   p->pend - p->ptr,
-                   0,
-                   &(p->regi));
+    pbeg = CURPTR(p);
+    plen = S_LEN(p) - p->idx;
+
+    CLEAR_MATCH_STATUS(p);
+    if (headonly) {
+        ret = re_match(RREGEXP(regex)->ptr,
+                       pbeg, plen,
+                       0,
+                       &(p->regi));
+    }
+    else {
+        ret = re_search(RREGEXP(regex)->ptr,
+                        pbeg, plen,
+                        0,
+                        plen,
+                        &(p->regi));
+    }
+
     if (ret == -2) {
         rb_raise(ScanError, "regexp buffer overflow");
+        return Qnil; /* not reach */
     }
-    else if ( ret < 0 ) {
+    else if (ret < 0) {
+        /* not matched */
         return Qnil;
     }
     else {
+        /* matched */
         size_t len = p->regi.end[0];
 
-        p->flags |= MATCH_FLAG;
-        if (ptrflag) {
-            p->pre = p->ptr;
-            p->ptr += len;
-            if (strflag)
-                return rb_str_new(p->pre, len);
+        SAVE_MATCH_STATUS(p, len);
+        if (succptr) {
+            p->idx += len;
+        }
+        if (getstr) {
+            return rb_str_new(pbeg, len);
         }
         else {
-            if (strflag)
-                return rb_str_new(p->ptr, len);
+            return INT2FIX(len);
         }
-        return INT2FIX(len);
     }
-
-    return Qnil; /* not reach */
 }
 
 static VALUE
 strscan_scan(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_scan(self, reg, 1, 1);
+    return strscan_do_scan(self, reg, 1, 1, 1);
 }
 
 static VALUE
 strscan_match_p(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_scan(self, reg, 0, 0);
+    return strscan_do_scan(self, reg, 0, 0, 1);
 }
 
 static VALUE
 strscan_skip(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_scan(self, reg, 1, 0);
+    return strscan_do_scan(self, reg, 1, 0, 1);
 }
 
 static VALUE
 strscan_check(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_scan(self, reg, 0, 1);
+    return strscan_do_scan(self, reg, 0, 1, 1);
 }
 
 static VALUE
 strscan_scan_full(self, reg, s, f)
 {
-    return strscan_do_scan(self, reg, RTEST(s), RTEST(f));
+    return strscan_do_scan(self, reg, RTEST(s), RTEST(f), 1);
 }
 
 
-static VALUE
-strscan_do_search(self, regex, ptrflag, strflag)
-    VALUE self, regex;
-    int ptrflag, strflag;
-{
-    struct strscanner *p;
-    int ret;
-
-    Check_Type(regex, T_REGEXP);
-    GET_SCANNER(self, p);
-
-    p->flags &= ~MATCH_FLAG;
-    ret = re_search(RREGEXP(regex)->ptr,
-                    p->ptr,
-                    p->pend - p->ptr,
-                    0,
-                    p->pend - p->ptr,
-                    &(p->regi));
-    if (ret == -2) {
-        rb_raise(ScanError, "regexp buffer overflow");
-    }
-    else if ( ret < 0 ) {
-        return Qnil;
-    }
-    else {
-        size_t len = p->regi.end[0];
-
-        p->flags |= MATCH_FLAG;
-        if (ptrflag) {
-            p->pre = p->ptr;
-            p->ptr += len;
-            if (strflag)
-                return rb_str_new(p->pre, len);
-        }
-        else {
-            if (strflag)
-                return rb_str_new(p->ptr, len);
-        }
-        return INT2FIX(len);
-    }
-
-    return Qnil; /* not reach */
-}
-
 static VALUE
 strscan_scan_until(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_search(self, reg, 1, 1);
+    return strscan_do_scan(self, reg, 1, 1, 0);
 }
 
 static VALUE
 strscan_exist_p(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_search(self, reg, 0, 0);
+    return strscan_do_scan(self, reg, 0, 0, 0);
 }
 
 static VALUE
 strscan_skip_until(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_search(self, reg, 1, 0);
+    return strscan_do_scan(self, reg, 1, 0, 0);
 }
 
 static VALUE
 strscan_check_until(self, reg)
     VALUE self, reg;
 {
-    return strscan_do_search(self, reg, 0, 1);
+    return strscan_do_scan(self, reg, 0, 1, 0);
 }
 
 static VALUE
 strscan_search_full(self, reg, s, f)
 {
-    return strscan_do_search(self, reg, RTEST(s), RTEST(f));
+    return strscan_do_scan(self, reg, RTEST(s), RTEST(f), 0);
 }
 
 
@@ -276,36 +300,56 @@ strscan_getch(self)
     VALUE self;
 {
     struct strscanner *p;
+    long len;
 
     GET_SCANNER(self, p);
-    p->flags &= ~MATCH_FLAG;
+    CLEAR_MATCH_STATUS(p);
     if (SCAN_FINISHED(p))
         return Qnil;
 
-    p->flags |= MATCH_FLAG;
-    p->pre = p->ptr;
-    p->ptr += mbclen(*p->ptr);
-    if (p->ptr > p->pend)
-        p->ptr = p->pend;
-    return rb_str_new(p->pre, p->ptr - p->pre);
+    len = mbclen(*CURPTR(p));
+    if (p->idx + len > S_LEN(p))
+        len = S_LEN(p) - p->idx;
+    SAVE_MATCH_STATUS(p, len);
+    p->idx += len;
+    return rb_str_new(PREPTR(p), p->idx - p->pre_beg);
 }
 
-
 static VALUE
-strscan_getbyte(self)
+strscan_get_byte(self)
     VALUE self;
 {
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    p->flags &= ~MATCH_FLAG;
+    CLEAR_MATCH_STATUS(p);
     if (SCAN_FINISHED(p))
         return Qnil;
 
-    p->flags |= MATCH_FLAG;
-    p->pre = p->ptr;
-    p->ptr++;
-    return rb_str_new(p->pre, 1);
+    SAVE_MATCH_STATUS(p, 1);
+    p->idx++;
+    return rb_str_new(PREPTR(p), 1);
+}
+
+
+static VALUE
+strscan_peek(self, vlen)
+    VALUE self, vlen;
+{
+    struct strscanner *p;
+    long len;
+
+    GET_SCANNER(self, p);
+
+    len = NUM2LONG(vlen);
+    if (SCAN_FINISHED(p)) {
+        return rb_str_new("", 0);
+    }
+    else {
+        if (p->idx + len > S_LEN(p))
+            len = S_LEN(p) - p->idx;
+        return rb_str_new(CURPTR(p), len);
+    }
 }
 
 
@@ -322,7 +366,6 @@ strscan_empty_p(self)
         return Qfalse;
 }
 
-
 static VALUE
 strscan_rest_p(self)
     VALUE self;
@@ -336,6 +379,36 @@ strscan_rest_p(self)
         return Qtrue;
 }
 
+static VALUE
+strscan_rest(self)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    if (SCAN_FINISHED(p))
+        return rb_str_new("", 0);
+    else
+        return rb_str_new(CURPTR(p), S_LEN(p) - p->idx);
+}
+
+static VALUE
+strscan_rest_size(self)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    if (SCAN_FINISHED(p)) {
+        return INT2FIX(0);
+    }
+    else {
+        long tmp;
+        tmp = S_LEN(p) - p->idx;
+        return INT2FIX(tmp);
+    }
+}
+
 
 static VALUE
 strscan_matched_p(self)
@@ -350,17 +423,28 @@ strscan_matched_p(self)
         return Qfalse;
 }
 
-
 static VALUE
-strscan_restsize(self)
+strscan_matched(self)
     VALUE self;
 {
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    return INT2FIX(p->pend - p->ptr);
+    if (! MATCHED_P(p)) return Qnil;
+    return rb_str_new(PREPTR(p), p->pre_end - p->pre_beg);
 }
 
+static VALUE
+strscan_matched_size(self)
+    VALUE self;
+{
+    struct strscanner *p;
+
+    GET_SCANNER(self, p);
+    if (! MATCHED_P(p)) return Qnil;
+
+    return INT2NUM(p->pre_end - p->pre_beg);
+}
 
 static VALUE
 strscan_unscan(self)
@@ -370,27 +454,13 @@ strscan_unscan(self)
 
     GET_SCANNER(self, p);
     if (! MATCHED_P(p))
-        rb_raise(ScanError, "cannot unscan: not scanned yet");
-    p->ptr = p->pre;
-    p->pre = 0;
-    return Qtrue;
+        rb_raise(ScanError, "cannot unscan: prev match is not valid");
+
+    p->idx = p->pre_beg;
+    CLEAR_MATCH_STATUS(p);
+    return self;
 }
 
-
-static VALUE
-strscan_matched(self)
-    VALUE self;
-{
-    struct strscanner *p;
-
-    GET_SCANNER(self, p);
-    if (MATCHED_P(p))
-        return rb_str_new(p->pre, p->ptr - p->pre);
-    else
-        return Qnil;
-}
-
-
 static VALUE
 strscan_pre_match(self)
     VALUE self;
@@ -398,9 +468,10 @@ strscan_pre_match(self)
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    return rb_str_new(p->pbeg, p->pre - p->pbeg);
-}
+    if (! MATCHED_P(p)) return Qnil;
 
+    return rb_str_new(S_PTR(p), p->pre_beg);
+}
 
 static VALUE
 strscan_post_match(self)
@@ -409,9 +480,10 @@ strscan_post_match(self)
     struct strscanner *p;
 
     GET_SCANNER(self, p);
-    return rb_str_new(p->ptr, p->pend - p->ptr);
-}
+    if (! MATCHED_P(p)) return Qnil;
 
+    return rb_str_new(S_PTR(p) + p->pre_end, S_LEN(p) - p->pre_end);
+}
 
 static VALUE
 strscan_aref(self, idx)
@@ -421,62 +493,30 @@ strscan_aref(self, idx)
     long i;
 
     GET_SCANNER(self, p);
-    if (! MATCHED_P(p))
-        return Qnil;
+    if (! MATCHED_P(p))        return Qnil;
     
     i = NUM2LONG(idx);
-    if (i >= p->regi.num_regs)
-        return Qnil;
-    
-    if (p->regi.beg[i] == -1)
-        return Qnil;
-    
-    return rb_str_new(p->pre + p->regi.beg[i],
+    if (i < 0)                 return Qnil;
+    if (i >= p->regi.num_regs) return Qnil;
+    if (p->regi.beg[i] == -1)  return Qnil;
+
+    return rb_str_new(PREPTR(p) + p->regi.beg[i],
                       p->regi.end[i] - p->regi.beg[i]);
 }
 
 
-static VALUE
-strscan_matchedsize(self)
-    VALUE self;
+static void
+cat_i_char(ret, c)
+    VALUE ret;
+    int c;
 {
-    struct strscanner *p;
+    char buf[1];
 
-    GET_SCANNER(self, p);
-    if (p->pre == 0)
-        return Qnil;
-    else
-        return INT2NUM(p->ptr - p->pre);
-}
-
-
-static VALUE
-strscan_string(self)
-    VALUE self;
-{
-    struct strscanner *p;
-
-    GET_SCANNER(self, p);
-    return rb_str_new(p->pbeg, p->pend - p->pbeg);
-}
-
-
-static VALUE
-strscan_peek(self, vlen)
-    VALUE self, vlen;
-{
-    struct strscanner *p;
-    long len;
-
-    GET_SCANNER(self, p);
-
-    len = NUM2LONG(vlen);
-    if (p->ptr + len > p->pend)
-        len = p->pend - p->ptr;
-
-    return rb_str_new(p->ptr, len);
+    buf[0] = c;
+    rb_str_cat(ret, buf, 1);
 }
 
+#define CLEN 5
 
 static VALUE
 strscan_inspect(self)
@@ -488,13 +528,44 @@ strscan_inspect(self)
     long len;
 
     GET_SCANNER(self, p);
-    ret = rb_str_new("#<StringScanner:", 16);
-    len = sprintf(buf, "0x%lx pos=%ld/%ld>",
-                  (unsigned long)p,
-                  (unsigned long)(p->ptr - p->pbeg),
-                  (unsigned long)(p->pend - p->pbeg));
-    rb_str_cat(ret, buf, len);
+    len = sprintf(buf, "#<%s %ld/%ld",
+                  rb_class2name(CLASS_OF(self)),
+                  p->idx, S_LEN(p));
+    ret = rb_str_new(buf, len);
+    
+    if (SCAN_FINISHED(p)) {
+        rb_str_cat(ret, " fin>", 4);
+    }
+    else {
+        char *sp;
 
+        sp = CURPTR(p) - CLEN;
+        if (sp < S_PTR(p)) sp = S_PTR(p);
+        if (sp != CURPTR(p)) {
+            rb_str_cat(ret, " \"", 2);
+            if (sp > S_PTR(p))
+                rb_str_cat(ret, "...", 3);
+            for (; sp < CURPTR(p); sp++) {
+                cat_i_char(ret, *sp);
+            }
+            rb_str_cat(ret, "\"", 1);
+        }
+        rb_str_cat(ret, " @", 2);
+        if (sp != S_END(p)) {
+            char *e;
+
+            e = sp + CLEN;
+            if (e > S_END(p)) e = S_END(p);
+            rb_str_cat(ret, " \"", 2);
+            for (; sp < e; sp++) {
+                cat_i_char(ret, *sp);
+            }
+            if (sp < S_END(p))
+                rb_str_cat(ret, "...", 3);
+            rb_str_cat(ret, "\"", 1);
+        }
+        rb_str_cat(ret, ">", 1);
+    }
     return ret;
 }
 
@@ -502,12 +573,18 @@ strscan_inspect(self)
 void
 Init_strscan()
 {
-    ScanError = rb_define_class("ScanError", rb_eStandardError);
-    StringScanner = rb_define_class("StringScanner", rb_cObject);
+    StringScanner = rb_define_class("C_StringScanner", rb_cObject);
+    rb_const_set(StringScanner,
+                 rb_intern("Version"), rb_str_new2(STRSCAN_VERSION));
 
     rb_define_singleton_method(StringScanner, "new", strscan_s_new, -1);
     rb_define_method(StringScanner, "reset",       strscan_reset,       0);
-    rb_define_method(StringScanner, "clear",       strscan_clear,       0);
+    rb_define_method(StringScanner, "terminate",   strscan_terminate,   0);
+    rb_define_method(StringScanner, "clear",       strscan_terminate,   0);
+    rb_define_method(StringScanner, "string",      strscan_get_string,  0);
+    rb_define_method(StringScanner, "string=",     strscan_set_string,  1);
+    rb_define_method(StringScanner, "pointer",     strscan_get_pointer, 0);
+    rb_define_method(StringScanner, "pointer=",    strscan_set_pointer, 1);
 
     rb_define_method(StringScanner, "scan",        strscan_scan,        1);
     rb_define_method(StringScanner, "skip",        strscan_skip,        1);
@@ -522,20 +599,25 @@ Init_strscan()
     rb_define_method(StringScanner, "search_full", strscan_search_full, 3);
 
     rb_define_method(StringScanner, "getch",       strscan_getch,       0);
-    rb_define_method(StringScanner, "getbyte",     strscan_getbyte,     0);
-    rb_define_method(StringScanner, "empty?",      strscan_empty_p,     0);
-    rb_define_method(StringScanner, "rest?",       strscan_rest_p,      0);
-    rb_define_method(StringScanner, "unscan",      strscan_unscan,      0);
-    rb_define_method(StringScanner, "matched?",    strscan_matched_p,   0);
-    rb_define_method(StringScanner, "matched",     strscan_matched,     0);
-    rb_define_method(StringScanner, "matchedsize", strscan_matchedsize, 0);
-    rb_define_method(StringScanner, "pre_match",   strscan_pre_match,   0);
-    rb_define_method(StringScanner, "post_match",  strscan_post_match,  0);
-    rb_define_method(StringScanner, "rest",        strscan_post_match,  0);
-    rb_define_method(StringScanner, "restsize",    strscan_restsize,    0);
-    rb_define_method(StringScanner, "[]",          strscan_aref,        1);
-    rb_define_method(StringScanner, "string",      strscan_string,      0);
+    rb_define_method(StringScanner, "get_byte",    strscan_get_byte,    0);
+    rb_define_method(StringScanner, "getbyte",     strscan_get_byte,    0);
     rb_define_method(StringScanner, "peek",        strscan_peek,        1);
     rb_define_method(StringScanner, "peep",        strscan_peek,        1);
+
+    rb_define_method(StringScanner, "empty?",      strscan_empty_p,     0);
+    rb_define_method(StringScanner, "rest?",       strscan_rest_p,      0);
+    rb_define_method(StringScanner, "rest",        strscan_rest,        0);
+    rb_define_method(StringScanner, "rest_size",   strscan_rest_size,   0);
+    rb_define_method(StringScanner, "restsize",    strscan_rest_size,   0);
+
+    rb_define_method(StringScanner, "matched?",    strscan_matched_p,   0);
+    rb_define_method(StringScanner, "matched",     strscan_matched,     0);
+    rb_define_method(StringScanner, "matched_size", strscan_matched_size, 0);
+    rb_define_method(StringScanner, "matchedsize", strscan_matched_size, 0);
+    rb_define_method(StringScanner, "unscan",      strscan_unscan,      0);
+    rb_define_method(StringScanner, "pre_match",   strscan_pre_match,   0);
+    rb_define_method(StringScanner, "post_match",  strscan_post_match,  0);
+    rb_define_method(StringScanner, "[]",          strscan_aref,        1);
+
     rb_define_method(StringScanner, "inspect",     strscan_inspect,     0);
 }
