@@ -8,7 +8,6 @@
 #   the GNU Lesser General Public License version 2 or later.
 #
 
-require 'amstd/bug'
 require 'strscan'
 
 
@@ -57,7 +56,7 @@ class Scanner
       end
 
       if tmp = @scan.scan( qpair ) then
-        ret << (preserve ? tmp : tmp[1, tmp.size - 1])
+        ret << (preserve ? tmp : tmp[1..-1])
 
       elsif @scan.skip( termexp )
         break
@@ -84,7 +83,7 @@ class Scanner
   end
 
   def scan_bug!( mes = 'must not happen' )
-    bug! "#{@filename}:#{@lineno}: #{mes}"
+    raise "[Script Bug] #{msg}"
   end
 
   def debug_report( arr )
