@@ -18,7 +18,7 @@ module TMail
 
 class Mailp < Racc::Parser
 
-module_eval <<'..end mailp.y modeval..idc7df893a05', 'mailp.y', 292
+module_eval <<'..end mailp.y modeval..id597aecd12c', 'mailp.y', 293
 
   MAILP_DEBUG = false
 
@@ -131,9 +131,9 @@ module_eval <<'..end mailp.y modeval..idc7df893a05', 'mailp.y', 292
     end
   end
 
-..end mailp.y modeval..idc7df893a05
+..end mailp.y modeval..id597aecd12c
 
-##### racc 1.2.5 generates ###
+##### racc 1.3.5 generates ###
 
 racc_reduce_table = [
  0, 0, :racc_error,
@@ -951,7 +951,7 @@ module_eval <<'.,.,', 'mailp.y', 261
   result
  end
 
-end
+end   # class Mailp
 
 
 end   # module TMail
